#include <curses.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include "opts.h"
#include "version.h"

WINDOW *playfield, *score;
time_t start_time = 0;
int defwidth = 20, defheight = 20;
int defmines = 70;
int width = 20, height = 20;
int mines = 70;
int nmines, rnmines;
int curses_inited = 0;

typedef struct {
	int mine;
	int mcount;
	enum { UNKNOWN, FLAG, EMPTY, MCOUNT, MINE } show;
} spot;

spot **field;

char playchars[] = "-+ X*";

option opts[] = {
	{ OTYPE_INT, 'h', "height", "height of playfield", OFLAG_NONE,
		&height, &defheight, 0, 0 },
	{ OTYPE_INT, 'm', "mines", "mines of playfield", OFLAG_NONE,
		&mines, &defmines, 0, 0 },
	{ OTYPE_INT, 'w', "width", "width of playfield", OFLAG_NONE,
		&width, &defwidth, 0, 0 },
	{ OTYPE_NUL, 0, 0, 0, 0, 0, 0, 0 }
};

int rnd(int);
static void die(char *, ...);
void trapper(int);
void init_level(void), main_loop(void);
void draw_border(void), draw_playfield(void), dead(void), draw_score(void);
void show(int, int), update(int, int), reveal(int, int);

int
rnd(int n) {
	return n < 2 ? 0 : random() % n;
}

static void
die(char *s, ...)
{
	va_list ap;

	if (curses_inited) {
		nocbreak();
		echo();
		endwin();
	}
	fprintf(stderr, "fatal: ");
	va_start(ap, s);
	vfprintf(stderr, s, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

void
trapper(int signum) {
	die("handling signal %d\n", signum);
}

int
main(int argc, char *argv[]) {
	int i;

	initscr();
	signal(SIGINT, trapper);
	cbreak();
	noecho();
	curses_inited = 1;
	optsgets(argc, argv, opts);
	if (width * 2 + 10 > COLS) {
		die("Sorry, I can only support a width of %d on this screen.\n",
			(COLS - 10) / 2);
	}
	if (height + 2 > LINES) {
		die("Sorry, I can only support a height of %d on this screen.\n",
			LINES - 2);
	}
	if (height * width < (mines * 2)) {
		die("Sorry, I am unhappy about a playfield that's half mines.\n");
	}
	playfield = newwin(height, width * 2, 1, 1);
	score = newwin(3, 8, 1, COLS - 8);
	if (!playfield || !score) {
		die("Couldn't create play windows.\n");
	}
	field = malloc(height * sizeof(spot *));
	if (!field) {
		die("couldn't allocate rows\n");
	}

	for (i = 0; i < height; ++i) {
		field[i] = calloc(width, sizeof(spot));
		if (!field[i]) {
			die("couldn't allocate row %d\n", i);
		}
	}
	start_time = time(NULL);
	srandom((int) time(NULL) + (int) getpid);
	init_level();
	draw_border();
	draw_playfield();
	main_loop();
	sleep(1);
	endwin();
	for (i = 0; i < height; ++i) {
		free(field[i]);
	}
	free(field);
	return 0;
}

void
init_level() {
	int i, j;
	int x, y;
	for (nmines = 0; nmines < mines; ++nmines) {
		x = rnd(width);
		y = rnd(height);
		if (field[y][x].mine) {
			--nmines;
		} else {
			field[y][x].mine = 1;
			for (i = y - 1; i < y + 2; ++i) {
				for (j = x - 1; j < x + 2; ++j) {
					if (i >= 0 && i < height && j >= 0 && j < width)
						++field[i][j].mcount;
				}
			}
		}
	}
	rnmines = nmines;
}

void
update(int y, int x) {
	wmove(playfield, y, x * 2);
	if (field[y][x].show == MCOUNT) {
		waddch(playfield, '0' + field[y][x].mcount);
	} else {
		waddch(playfield, playchars[field[y][x].show]);
	}

}

void
show(int y, int x) {
	int i, j;
	if (field[y][x].show != FLAG && field[y][x].mine)
		dead();
	if (field[y][x].show)
		return;
	switch (field[y][x].mcount) {
	case 0:
		field[y][x].show = EMPTY;
		for (i = y - 1; i < y + 2; ++i) {
			for (j = x - 1; j < x + 2; ++j) {
				if (i >= 0 && i < height && j >= 0 && j < width)
					show(i, j);
			}
		}
		break;
	default:
		field[y][x].show = MCOUNT;
		break;
	}
	update(y, x);
}

void
reveal(int y, int x) {
	int i, j;
	int count = 0;
	for (i = y - 1; i < y + 2; ++i) {
		for (j = x - 1; j < x + 2; ++j) {
			if (i >= 0 && i < height && j >= 0 && j < width)
				if (field[i][j].show == FLAG)
					++count;
		}
	}
	if (count == field[y][x].mcount) {
		for (i = y - 1; i < y + 2; ++i) {
			for (j = x - 1; j < x + 2; ++j) {
				if (i >= 0 && i < height && j >= 0 && j < width)
					show(i, j);
			}
		}
	}
	wrefresh(playfield);
}

void
main_loop() {
	int xloc = width / 2;
	int yloc = height / 2;
	int c;

	while (nmines > 0) {
		draw_score();
		wmove(playfield, yloc, xloc * 2);
		wrefresh(playfield);
		c = getch();
		switch (c) {
		case 'h':
			--xloc;
			break;
		case 'j':
			++yloc;
			break;
		case 'k':
			--yloc;
			break;
		case 'l':
			++xloc;
			break;
		case ' ':
			if (field[yloc][xloc].mine) {
				dead();
			} else {
				show(yloc, xloc);
				wrefresh(playfield);
			}
			break;
		case '\t':
			reveal(yloc, xloc);
			break;
		case '!':
			if (field[yloc][xloc].show == FLAG) {
				field[yloc][xloc].show = UNKNOWN;
				++nmines;
				if (field[yloc][xloc].mine)
					++rnmines;
			} else if (field[yloc][xloc].show == UNKNOWN) {
				field[yloc][xloc].show = FLAG;
				if (field[yloc][xloc].mine)
					--rnmines;
				--nmines;
			}
			update(yloc, xloc);
			wrefresh(playfield);
			break;
		case 'q':
			nmines = -1;
			break;
		}
	}
	if (rnmines) {
		nmines = -1;
	}
	draw_score();
}

void
draw_border() {
	int i;

	wmove(stdscr, 0, 0);
	waddch(stdscr, '+');
	for (i = 1; i < width * 2; ++i) {
		waddch(stdscr, '-');
	}
	waddch(stdscr, '+');
	for (i = 1; i < height + 1; ++i) {
		wmove(stdscr, i, 0);
		waddch(stdscr, '|');
		wmove(stdscr, i, width * 2);
		waddch(stdscr, '|');
	}
	wmove(stdscr, height + 1, 0);
	waddch(stdscr, '+');
	for (i = 1; i < width * 2; ++i) {
		waddch(stdscr, '-');
	}
	waddch(stdscr, '+');
	refresh();
}

void
draw_playfield() {
	int i, j;
	for (i = 0; i < height; ++i) {
		for (j = 0; j < width; ++j) {
			update(i, j);
		}
	}
	wrefresh(playfield);

}

void
draw_score(void) {
	wclear(score);
	wmove(score, 0, 0);
	wprintw(score, "T:  %4d\n", (int) difftime(time(NULL), start_time));
	wprintw(score, "M:  %4d", nmines);
	wrefresh(score);
}

void
dead(void) {
	int i, j;
	for (i = 0; i < height; ++i)
		for (j = 0; j < width; ++j)
			if (field[i][j].mine && field[i][j].show != FLAG) {
				field[i][j].show = MINE;
				update(i, j);
			}
	wrefresh(playfield);
	die("you hit a mine.\n");
}
